import {MemoryManager, Byte} from "./MemoryManager";
import {Compiler} from "./Complier.ts";

export class RunningError extends Error {
    constructor(message: string, index: number) {
        super(message);
        this.index = index;
    }

    index: number;
}

export enum RunnerState {
    Stop,
    Running,
    Step
}

export class Runner {
    // 与Compiler 类似的事件机制 要是再来一个我就把这段逻辑单独抽出来
    //[PCChanged,RegistersChanged,RunningError(string,RunningError)]
    _events: Map<string, ((arg0: string, arg1?: any) => void)[]> = new Map<string, ((arg0: string, arg1?: any) => void)[]>();

    AddEventListener(event: string, callback: (arg0: string, arg1?: any) => void) {
        if (!this._events.has(event))
            this._events.set(event, [callback]);
        else
            this._events.get(event)!.push(callback);
    }

    Notify(event: string, data?: any) {
        if (this._events.has(event))
            if (data === undefined)
                this._events.get(event)!.forEach(callback => callback(event));
            else
                this._events.get(event)!.forEach(callback => callback(event, data));
    }

    State: RunnerState = RunnerState.Stop;
    _runningSpeed: number = 20;
    _processor: Processor;
    _memory: MemoryManager;
    _processCounter: number = 0;
    _compiler: Compiler;
    _stack: [Byte[], number][] = [];
    _intervalId: NodeJS.Timeout | null = null;

    set RunningSpeed(speed: number) {
        this._runningSpeed = speed;
        if (this.State === RunnerState.Running) {
            this.Pause();
            this.Resume();
        }
    }

    get RunningSpeed() {
        return this._runningSpeed;
    }

    StartInterval() {
        if (this._intervalId === null)
            this._intervalId = setInterval(() => {
                if (this._processCounter >= this._compiler.PCToCode.size) {
                    this.Stop();
                    return;
                }
                this.RunOneLineCode();
            }, this._runningSpeed);
    }

    StopInterval() {
        if (this._intervalId !== null) {
            clearInterval(this._intervalId);
            this._intervalId = null;
        }
    }

    constructor(memoryManager: MemoryManager, compiler: Compiler) {
        this._memory = memoryManager
        this._processor = new Processor();
        this._compiler = compiler;
        compiler.AddEventListener("CompileFinished", (_) => {
            this._memory.SetMemSet(compiler.MemorySet)
        });
        // Bind methods to the instance
        this.GetMemory = this.GetMemory.bind(this);
        this.SetMemory = this.SetMemory.bind(this);
        this.SetPC = this.SetPC.bind(this);
        this.SaveEnv = this.SaveEnv.bind(this);
        this.RestoreEnv = this.RestoreEnv.bind(this);
    }

    CommandParsing(): [string, number | string, number | string] {
        let code = "";
        if (this._compiler.PCToCode.has(this._processCounter))
            code = this._compiler.PCToCode.get(this._processCounter) as string;
        else
            throw new RunningError("PC out of range", 0);
        this._processCounter++;
        let mnemonicAndVals = code.split(/[\s,]+/);
        // 如果指令不合法 Compiler应该会有异常抛出 则不能执行到这里
        if (mnemonicAndVals.length !== 3)
            throw new RunningError("Invalid code", this._processCounter);
        let mnemonic = mnemonicAndVals[0];

        function getVal(runner: Runner, val: string) {
            if (runner._processor._registers.has(val))
                return val;
            else if (runner._compiler.Labels.has(val))
                return runner._compiler.Labels.get(val) as number;
            else {
                let t = parseInt(val);
                if (isNaN(t)) {
                    throw new RunningError("Invalid argument", runner._processCounter);
                }
                return t;
            }
        }

        return [mnemonic, getVal(this, mnemonicAndVals[1]), getVal(this, mnemonicAndVals[2])];
    }

    RunOneLineCode() {
        try {
            this._processor.RunCode(...this.CommandParsing(), this.GetMemory, this.SetMemory, this.SetPC, this.SaveEnv, this.RestoreEnv);
        }catch (e:any)
        {
            console.log(e);
            e = e as RunningError;
            if(e === undefined)
                throw new RunningError("Unknown error",this._processCounter);
            e.index =  e.index === -1? this._processCounter : e.index;
            this.Pause();
            this.Notify("RunningError", e);
        }
        this.Notify("PCChanged");
        this.Notify("RegistersChanged");
    }

    Init() {
        this._memory.clear();
        this._compiler.CompileCode();
        this._processCounter = 0;
    }

    Run() {
        this.State = RunnerState.Running;
        this.Init();
        this.StartInterval();
    }

    Resume() {
        this.State = RunnerState.Running;
        this.StartInterval();
    }

    Step() {
        if (this._processCounter >= this._compiler.PCToCode.size)
            return;
        if (this.State === RunnerState.Running) {
            this.Stop();
        }
        this.State = RunnerState.Step;
        this.RunOneLineCode()
    }

    Pause() {
        this.State = RunnerState.Stop;
        this.StopInterval()
    }

    Stop() {
        this.State = RunnerState.Stop;
        this.StopInterval()
        this._processCounter = 0;
        this._memory.clear();
    }

    Refresh() {
        this.Run();
    }

    GetMemory(t: number): Byte {
        return this._memory.GetVal(t);
    }

    SetMemory(t: number, val: Byte): void {
        this._memory.SetVal(t, val);
    }

    SaveEnv() {
        if (this._stack.length > 8)
            throw new RunningError("Stack overflow",this._processCounter);
        this._stack.push([this._processor.GetRegisters().map(x => x.Copy()), this._processCounter]);
    }

    RestoreEnv() {
        if (this._stack.length === 0) {
            this.Pause();
            return;
        }
        let [regs, pc] = this._stack.pop()!;
        this._processor.SetRegisters(regs);
        this._processCounter = pc;
    }

    SetPC(addr: number) {
        this._processCounter = addr;
    }

    GetRegisters(): Byte[] {
        return this._processor.GetRegisters();
    }

    GetPC(): number {
        return this._processCounter;
    }
}

class Processor {
    _registers: Map<string, Byte>;

    constructor() {
        this._registers = new Map<string, Byte>([
            ["AX", new Byte()],
            ["BX", new Byte()],
            ["CX", new Byte()],
            ["DX", new Byte()],
            ["EX", new Byte()],
            ["FX", new Byte()],
            ["GX", new Byte()]
        ]);
    }

// static KeyWords = ["nop","ldc","add","div","and","or","xor","not","lsl","lsr","ld","st","mov","cp","jp","ltjp","gtjp","eqjp","call","ret"];
// static Registers = ["AX","BX","CX","DX","EX","FX","GX"];
    GetRegister(register: string | number): Byte {
        if (this._registers.has(register as string))
            return this._registers.get(register as string)!;
        else
            throw new RunningError("Invalid register",-1);
    }

    GetRegisters(): Byte[] {
        return Array.from(this._registers.values());
    }

    SetRegisters(registers: Byte[]) {
        for (let i = 0; i < 7; i++) {
            this._registers.get(String.fromCharCode('A'.charCodeAt(0) + i) + 'X')!.SetValue(registers[i]);
        }
    }

    GetRegisterValue(register: string | number): number {
        return this.GetRegister(register).GetValue() as number;
    }

    SetRegisterValue(register: string | number, value: number) {
        try {
            this.GetRegister(register).SetValue(value);
        }catch (e)
        {
            throw new RunningError("Invalid value",-1);
        }
    }

    InRegister(register: string | number): boolean {
        return typeof register === 'string' && this._registers.has(register);
    }

    RunCode(mnemonic: string, operandA: number | string, operandB: number | string, getMemory: (addr: number) => Byte, setMemory: (addr: number, value: Byte) => void, setPC: (addr: number) => void, saveEnv: () => void, restoreEnv: () => void) {
        {
            let a = operandA;
            let b = operandB;
            if (this.InRegister(a))
                a = a + '(' + this.GetRegisterValue(a) + ')';
            if (this.InRegister(b))
                b = b + '(' + this.GetRegisterValue(b) + ')';
            console.log(mnemonic,a,b);
        }
        switch (mnemonic) {
            case "nop":
                break;
            case "ldc":
                if (this.InRegister(operandA) && typeof operandB === 'number') {
                    this.GetRegister(operandA).SetValue(operandB);
                    return;
                }
                break;
            case "add":
                if (this.InRegister(operandA) && this.InRegister(operandB)) {
                    this.GetRegister(operandA).AddAndAssign(this._registers.get(operandB as string)!);
                    return;
                }
                break;
            case "div":
                if (this.InRegister(operandA) && this.InRegister(operandB)) {
                    let a = this.GetRegisterValue(operandA), b = this.GetRegisterValue(operandB);
                    this.SetRegisterValue(operandA, Math.floor(a / b));
                    this.SetRegisterValue("AX", a % b);
                    return;
                }
                break;
            case "and":
                if (this.InRegister(operandA) && this.InRegister(operandB)) {
                    this.GetRegister(operandA).AndAndAssign(this.GetRegister(operandB));
                    return;
                }
                break;
            case "or":
                if (this.InRegister(operandA) && this.InRegister(operandB)) {
                    this.GetRegister(operandA).OrAndAssign(this.GetRegister(operandB));
                    return;
                }
                break;
            case "xor":
                if (this.InRegister(operandA) && this.InRegister(operandB)) {
                    this.GetRegister(operandA).XorAndAssign(this.GetRegister(operandB));
                    return;
                }
                break;
            case "not":
                if (this.InRegister(operandA) && operandB === 0) {
                    this.GetRegister(operandA).NotAndAssign();
                    return;
                }
                break;
            case "lsl":
                if (this.InRegister(operandA) && this.InRegister(operandB)) {
                    this.GetRegister(operandA).LslAndAssign(this.GetRegister(operandB));
                    return;
                }
                break;
            case "lsr":
                if (this.InRegister(operandA) && this.InRegister(operandB)) {
                    this.GetRegister(operandA).LsrAndAssign(this.GetRegister(operandB));
                    return;
                }
                break;
            case "ld":
                if (this.InRegister(operandA) && typeof operandB === 'number') {
                    this.GetRegister(operandA).SetValue(getMemory(operandB as number));
                    return;
                }
                break;
            case "st":
                if (this.InRegister(operandA) && typeof operandB === 'number') {
                    setMemory(operandB as number, this.GetRegister(operandA));
                    return;
                }
                break;
            case "mov":
                if (this.InRegister(operandA) && this.InRegister(operandB)) {
                    this.GetRegister(operandA).SetValue(this.GetRegister(operandB));
                    return;
                } else if (this.InRegister(operandA) && typeof operandB === 'number') {
                    this.GetRegister(operandA).SetValue(operandB);
                    return;
                }
                break;
            case "cp":
                operandA = this.InRegister(operandA) ? this.GetRegisterValue(operandA) : operandA;
                operandB = this.InRegister(operandB) ? this.GetRegisterValue(operandB) : operandB;
                if (typeof operandA === 'number' && typeof operandB === 'number') {
                    for (let i = 0; i < (this._registers.get("BX")!.GetValue() as number); i++)
                        setMemory(operandB + i, getMemory(operandA + i));
                    return;
                }
                break;
            case "jp":
                if (typeof operandA === 'number' && typeof operandB === 'number' && operandA === 0) {
                    setPC(operandB as number);
                    return;
                }
                break;
            case "ltjp":
                if (this.InRegister(operandA) && typeof operandB === 'number') {
                    if (this.GetRegisterValue("BX") < this.GetRegisterValue(operandA))
                        setPC(operandB as number);
                    return;
                }
                break;
            case "gtjp":
                if (this.InRegister(operandA) && typeof operandB === 'number') {
                    if (this.GetRegisterValue("BX") > this.GetRegisterValue(operandA))
                        setPC(operandB as number);
                    return;
                }
                break;
            case "eqjp":
                if (this.InRegister(operandA) && typeof operandB === 'number') {
                    if (this.GetRegisterValue("BX") === this.GetRegisterValue(operandA))
                        setPC(operandB as number);
                    return;
                }
                break;
            case "call":
                if (typeof operandA === 'number' && typeof operandB === 'number' && operandA === 0) {
                    saveEnv();
                    setPC(operandB as number);
                    return;
                }
                break;
            case "ret":
                if (typeof operandA === 'number' && typeof operandB === 'number' && operandA === 0 && operandB === 0) {
                    restoreEnv();
                    return;
                }
                break
            default:
                throw new RunningError("Invalid mnemonic",-1);
        }
        throw new RunningError("Invalid argument",-1);
    }
}
